/*-------------------------------------------------------------------------
 *  vmadump.c:  Virtual Memory Area dump/restore routines
 *
 *  Copyright (C) 1999-2001 by Erik Hendriks <erik@hendriks.cx>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: vmadump.c,v 1.7 2002/12/09 13:38:15 eugen Exp $
 *-----------------------------------------------------------------------*/
#define EXPORT_SYMTAB

#include <linux/config.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif
#if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/errno.h>
#include <linux/binfmts.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <linux/personality.h>
#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#ifdef __i386__
#include <asm/i387.h>
#endif
#ifdef __sparc__
#include <asm/cprefix.h>
#endif

#include "vmadump.h"
#ifdef __WITH__CHPOX__
#include "chpox.h"
#endif

#ifdef EXTENDED_MUNMAP
#define vmad_do_munmap(a,b,c) do_munmap((a),(b),(c),1)
#else
#define vmad_do_munmap do_munmap
#endif

static char vmad_magic[3] = VMAD_MAGIC;
extern void * sys_call_table[];
static int syscall_num = 0;

MODULE_AUTHOR("Erik Hendriks <erik@hendriks.cx>");
MODULE_DESCRIPTION("VMADump - Virtual Memory Area Dumper");
MODULE_LICENSE("GPL");
MODULE_PARM(syscall_num, "i");
MODULE_PARM_DESC(syscall_num,
"Syscall number to allow calling VMADump directly.  The default (0) means "
"no VMADump syscall.  There is no bounds check on the syscall number so "
"be careful with this option.");

/*--------------------------------------------------------------------
 *  Some utility stuff for reading and writing and misc kernel syscalls.
 *------------------------------------------------------------------*/
static
ssize_t default_read_file(struct vmadump_map_ctx *ctx, struct file *file,
			  void *buf, size_t count) {
    ssize_t ret = -EBADF;
    if (file->f_mode & FMODE_READ) {
	ret = locks_verify_area(FLOCK_VERIFY_READ, file->f_dentry->d_inode,
				file, file->f_pos, count);
	if (!ret) {
	    ssize_t (*read)(struct file *, char *, size_t, loff_t *);
	    ret = -EINVAL;
	    if (file->f_op && (read = file->f_op->read) != NULL)
		ret = read(file, buf, count, &file->f_pos);
	}
    }
    return ret;
}

static
ssize_t read_user(struct vmadump_map_ctx *ctx,
		  struct file *file, void *buf, size_t count) {
    ssize_t r, bytes = count;
    ssize_t (*rfunc)(struct vmadump_map_ctx *ctx, struct file *file,
		     void *buf, size_t count);
    rfunc = (ctx && ctx->read_file) ? ctx->read_file : default_read_file;
    while (bytes) {
	r = rfunc(ctx, file, buf, bytes);
	if (r < 0)  return r;
	if (r == 0) return count - bytes;
	bytes -= r; buf += r;
    }
    return count;
}

static
ssize_t read_kern(struct vmadump_map_ctx *ctx, 
		  struct file *file, void *buf, size_t count) {
    ssize_t err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = read_user(ctx, file, buf, count);
    set_fs(oldfs);
    return err;
}

static
ssize_t default_write_file(struct vmadump_map_ctx *ctx, struct file *file,
			   const void *buf, size_t count) {
    ssize_t ret = -EBADF;
    if (file->f_mode & FMODE_WRITE) {
	struct inode *inode = file->f_dentry->d_inode;
	ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file,
				file->f_pos, count);
	if (!ret) {
	    ssize_t (*write)(struct file *, const char *, size_t, loff_t *);
	    ret = -EINVAL;
	    if (file->f_op && (write = file->f_op->write) != NULL)
		ret = write(file, buf, count, &file->f_pos);
	}
    }
    return ret;
}

static
ssize_t write_user(struct vmadump_map_ctx *ctx, struct file *file,
		   const void *buf, size_t count) {
    ssize_t w, bytes = count;
    ssize_t (*wfunc)(struct vmadump_map_ctx *ctx, struct file *file,
		     const void *buf, size_t count);
    wfunc = (ctx && ctx->write_file) ? ctx->write_file : default_write_file;
    while (bytes) {
	w = wfunc(ctx, file, buf, bytes);
	if (w < 0)  return w;
	if (w == 0) return count - bytes;
	bytes -= w; buf += w;
    }
    return count;
}

static
ssize_t write_kern(struct vmadump_map_ctx *ctx, struct file *file,
		   const void *buf, size_t count) {
    ssize_t err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = write_user(ctx, file, buf, count);
    set_fs(oldfs);
    return err;
}

extern void *sys_call_table[];
static
long k_mprotect(unsigned long start, size_t len, unsigned long prot) {
    return ((long (*)(unsigned long,size_t,unsigned long))
	    sys_call_table[__NR_mprotect])(start,len,prot);
}

static
long k_exit(int code) {
    return ((long(*)(int))sys_call_table[__NR_exit])(code);
}

static
long k_access(const char *path, int mode) {
    return ((long(*)(const char *, int))
	    sys_call_table[__NR_access])(path,mode);
}

/*--------------------------------------------------------------------
 *  Library list stuff.
 *------------------------------------------------------------------*/
/* XXX This should be more configurable than this - this is idea of
 *     figuring out what's a library  */
struct liblist_entry {
    struct list_head list;
    char libname[0];
};

static LIST_HEAD(liblist);

static
void liblist_clear(void) {
    struct liblist_entry *entry;
    while (!list_empty(&liblist)) {
	entry = list_entry(liblist.next, struct liblist_entry, list);
	list_del(liblist.next);
	kfree(entry);
    }
}

static
struct liblist_entry *
liblist_find(const char *filename) {
    struct list_head *p;
    struct liblist_entry *entry;
    p = liblist.next;
    while (p != &liblist) {
	entry = list_entry(p, struct liblist_entry, list);
	if (strcmp(entry->libname, filename) == 0)
	    return entry;
	p = p->next;
    }
    return 0;
}

static
int liblist_add(char *filename) {
    struct liblist_entry *entry;

    /* Don't add things twice */
    if (liblist_find(filename)) return 0;

    entry = kmalloc(sizeof(*entry)+strlen(filename)+1, GFP_KERNEL);
    if (!entry) return -ENOMEM;
    strcpy(entry->libname, filename);
    list_add(&entry->list, &liblist);
    return 0;
}

static
int liblist_del(char *filename) {
    struct liblist_entry *entry;
    entry = liblist_find(filename);
    if (entry) {
	list_del(&entry->list);
	kfree(entry);
	return 0;
    }
    return -ENOENT;
}

/* Returns the size of our library list converted to text */
static
int liblist_size(void) {
    int len;
    struct list_head *p;
    struct liblist_entry *entry;

    len = 0;
    p = liblist.next;
    while (p != &liblist) {
	entry = list_entry(p, struct liblist_entry, list);
	len += strlen(entry->libname)+1;
	p = p->next;
    }
    len++;			/* for trailing null. */
    return len;
}

static
int do_lib_op(int request, char *buf, int size) {
    int err, len;
    char *filename;
    struct list_head *p;
    struct liblist_entry *entry;

    switch(request) {
    case VMAD_LIB_CLEAR:
	if (!suser()) return -EPERM;
	liblist_clear();
	return 0;

    case VMAD_LIB_ADD:
	if (!suser()) return -EPERM;
	filename = getname(buf);
	if (IS_ERR(filename)) return PTR_ERR(filename);
	err = liblist_add(filename);
	putname(filename);
	return err;

    case VMAD_LIB_DEL:
	if (!suser()) return -EPERM;
	filename = getname(buf);
	if (IS_ERR(filename)) return PTR_ERR(filename);
	err = liblist_del(filename);
	putname(filename);
	return err;

    case VMAD_LIB_SIZE:
	return liblist_size();

    case VMAD_LIB_LIST:
	len = liblist_size();
	if (len > size) return -ENOSPC;
	size = len;
	/* Copy all those strings out to user space. */
	p = liblist.next;
	while (p != &liblist) {
	    entry = list_entry(p, struct liblist_entry, list);
	    len = strlen(entry->libname);
	    if (copy_to_user(buf, entry->libname, len)) return EFAULT;
	    buf += len;
	    put_user('\0', buf++);
	    p = p->next;
	}
	put_user('\0', buf);
	return size;

    default:
	return -EINVAL;
    }
}

static
char *default_map_name(struct file *f, char *buffer, int size) {
    return d_path(f->f_dentry, f->f_vfsmnt, buffer, PAGE_SIZE);
}


/* this is gonna be handled with contexts too */
static
int is_library(const char *filename) {
    return (liblist_find(filename) != 0);
}


static
struct file *default_map_open(const char *filename) {
    struct file *file;
    mm_segment_t user_fs;
    user_fs = get_fs(); set_fs(KERNEL_DS);
    file = filp_open(filename,O_RDONLY,0);
    set_fs(user_fs);
    return file;
}

/*--------------------------------------------------------------------
 *  Dump hook stuff.
 *------------------------------------------------------------------*/
struct hook_t {
    struct list_head list;
    struct vmadump_hook *hook;
};
struct vmadump_hook_handle {
    int rw;			/* for checking on the user */
    struct vmadump_map_ctx *ctx;
    struct file *file;
};

static struct rw_semaphore hook_lock;
static LIST_HEAD(hooks);

int vmadump_add_hook(struct vmadump_hook *hook) {
    struct hook_t *h, *new_h;
    struct list_head *l;

    new_h = kmalloc(sizeof(*new_h), GFP_KERNEL);
    new_h->hook = hook;

    down_write(&hook_lock);
    /* check to make sure that this hook isn't being added twice */
    for (l = hooks.next; l != &hooks; l = l->next) {
	h = list_entry(l, struct hook_t, list);
	if (h->hook == hook) {
	    up_write(&hook_lock);
	    kfree(new_h);
	    return -EEXIST;
	}
    }
    list_add_tail(&new_h->list, &hooks);
    up_write(&hook_lock);
    printk(KERN_INFO "vmadump: Registered hook \"%s\"\n", hook->tag);
    return 0;
}

int vmadump_del_hook(struct vmadump_hook *hook) {
    struct hook_t *h;
    struct list_head *l;

    down_write(&hook_lock);
    for (l = hooks.next; l != &hooks; l = l->next) {
	h = list_entry(l, struct hook_t, list);
	
	if (h->hook == hook) {
	    list_del(&h->list);
	    up_write(&hook_lock);
	    kfree(h);
	    printk(KERN_INFO "vmadump: Unregistered hook \"%s\"\n", hook->tag);
	    return 0;
	}
    }
    up_write(&hook_lock);
    return -ENOENT;
}

struct vmadump_callback_handle {
    int rw;
    struct vmadump_map_ctx *ctx;
    struct file *file;
};

/* Call every hook freeze function */
static
int do_freeze_hooks(struct vmadump_map_ctx *ctx, struct file *file,
		    struct pt_regs *regs) {
    long bytes = 0, r;
    struct hook_t *h;
    struct list_head *l;
    struct vmadump_hook_header hookhdr;
    struct vmadump_hook_handle hh = { 1, ctx, file};

    down_read(&hook_lock);
    for (l = hooks.next; l != &hooks; l = l->next) {
	h = list_entry(l, struct hook_t, list);
	r = h->hook->freeze(&hh, regs);
	if (r < 0) {
	    up_read(&hook_lock);
	    return r;
	}
	bytes += r;
    }
    up_read(&hook_lock);

    /* Terminate the list of hooks */
    memset(&hookhdr, 0, sizeof(hookhdr));
    r = write_kern(ctx, file, &hookhdr, sizeof(hookhdr));
    if (r < 0) return r;
    if (r != sizeof(hookhdr)) return -EIO;
    bytes += r;

    return bytes;
}

static
long skip_data(struct vmadump_map_ctx *ctx, struct file *file, long len) {
    long r = 0;
    void *page;
    page = (void *) get_free_page(GFP_KERNEL);
    if (!page)
	return -ENOMEM;

    while (len > 0) {
	r = read_kern(ctx, file, page, (len>PAGE_SIZE) ? PAGE_SIZE : len);
	if (r <= 0) break;
	len -= r;
    }
    free_page((long) page);

    if (r == 0) r = -EIO;	/* end of file.... */
    return 0;
}

static
int do_thaw_hooks(struct vmadump_map_ctx *ctx, struct file *file,
		  struct pt_regs *regs) {
    long r;
    struct hook_t *h;
    struct list_head *l;
    struct vmadump_hook_header hookhdr;
    struct vmadump_hook_handle hh = { 0, ctx, file};

    r = read_kern(ctx, file, &hookhdr, sizeof(hookhdr));
    if (r != sizeof(hookhdr)) goto err;
    while (hookhdr.tag[0]) {
	/* Do a bit of sanity checking on this dump header */
	hookhdr.tag[VMAD_HOOK_TAG_LEN-1] = 0; /* null terminate that string... */
	if (hookhdr.size <= 0) {
	    r = -EINVAL;
	    goto err;
	}

	/* See if we find a matching hook */
	down_read(&hook_lock);
	for (l = hooks.next; l != &hooks; l = l->next) {
	    h = list_entry(l, struct hook_t, list);
	    if (strcmp(hookhdr.tag, h->hook->tag) == 0) {
		r = h->hook->thaw(&hh, &hookhdr, regs);
		break;
	    }
	}
	if (l == &hooks)
	    r = skip_data(ctx, file, hookhdr.size);
	up_read(&hook_lock);
	if (r) goto err;
	
	r = read_kern(ctx, file, &hookhdr, sizeof(hookhdr));
	if (r != sizeof(hookhdr)) goto err;
    }
    return 0;

 err:
    if (r >= 0) r = -EIO;
    return r;
}

/* read/write calls for use by hooks */
ssize_t vmadump_read_u(struct vmadump_hook_handle *h, void *buf, size_t count){
    if (h->rw != 0) return -EINVAL;
    return read_user(h->ctx, h->file, buf, count);
}

ssize_t vmadump_read_k(struct vmadump_hook_handle *h, void *buf, size_t count){
    ssize_t err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = vmadump_read_u(h, buf, count);
    set_fs(oldfs);
    return err;

}

ssize_t vmadump_write_u(struct vmadump_hook_handle *h,
			const void *buf, size_t count) {
    if (h->rw != 1) return -EINVAL;
    return write_user(h->ctx, h->file, buf, count);
}

ssize_t vmadump_write_k(struct vmadump_hook_handle *h,
			const void *buf, size_t count) {
    ssize_t err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = vmadump_write_u(h, buf, count);
    set_fs(oldfs);
    return err;
}


/*--------------------------------------------------------------------
 *  Process "thawing" routines.
 *------------------------------------------------------------------*/
static
int mmap_file(struct vmadump_map_ctx *ctx,
	      struct vmadump_vma_header *head, char *filename,
	      unsigned long prot, unsigned long flags) {
    struct file *file;
    long retval = -1;

    /* This is a lot like open w/o a file descriptor */
    if (ctx && ctx->map_open)
	file = ctx->map_open(ctx, filename);
    else
	file = default_map_open(filename);
    if (IS_ERR(file))
	return PTR_ERR(file);

    down_write(&current->mm->mmap_sem);
    retval = do_mmap(file, head->start, head->end - head->start,
		     prot, flags, head->offset);
    up_write(&current->mm->mmap_sem);
    retval = (retval != head->start);
    if (retval) printk("do_mmap(<file>, %p, %p, ...) failed: %p\n",
		       (void *) head->start, (void *) head->end-head->start,
		       (void *) retval);
    fput(file);
    return retval;
}

static
int load_map(struct vmadump_map_ctx *ctx,
	     struct file *file, struct vmadump_vma_header *head) {
    int r;
    unsigned long mmap_prot, mmap_flags, addr;
    struct vmadump_page_header page;

    mmap_prot  = 0;
    mmap_flags = MAP_FIXED|MAP_PRIVATE;
    if (head->flags & VM_READ)  mmap_prot |= PROT_READ;
    if (head->flags & VM_WRITE) mmap_prot |= PROT_WRITE;
    if (head->flags & VM_EXEC)  mmap_prot |= PROT_EXEC;
    if (head->flags & VM_GROWSDOWN) mmap_flags |= MAP_GROWSDOWN;
    if (head->flags & VM_EXECUTABLE) mmap_flags |= MAP_EXECUTABLE;
    if (head->flags & VM_DENYWRITE) mmap_flags |= MAP_DENYWRITE;

    if (head->namelen) {
	char *filename;
	if (head->namelen > PAGE_SIZE) {
	    printk("vmadump: thaw: bogus namelen %d\n", (int) head->namelen);
	    return -EINVAL;
	}
	filename = kmalloc(head->namelen+1,GFP_KERNEL);
	r = read_kern(ctx, file, filename, head->namelen);
	if (r != head->namelen) goto err;
	filename[head->namelen] = 0;

	if (mmap_file(ctx, head, filename,
		      PROT_READ|PROT_WRITE|PROT_EXEC, mmap_flags)){
	    kfree(filename);
	    printk("vmadump: mmap failed: %s\n", filename);
	    return -1;
	}
	kfree(filename);
    } else {
	/* Load the data from the dump file */
	down_write(&current->mm->mmap_sem);
	addr = do_mmap(0, head->start, head->end - head->start,
		       PROT_READ|PROT_WRITE|PROT_EXEC, mmap_flags, 0);
	up_write(&current->mm->mmap_sem);
	if (addr != head->start) {
	    printk("do_mmap(0, %08lx, %08lx, ...) = 0x%08lx (failed)\n",
		   head->start, head->end - head->start, addr);
	    return -EINVAL;
	}
    }

    /* Read in patched pages */
    r = read_kern(ctx, file, &page, sizeof(page));
    while (r == sizeof(page) && page.start != ~0UL) {
	r = read_user(ctx, file, (void *) page.start, PAGE_SIZE);
	if (r != PAGE_SIZE) goto err;
	if (mmap_prot & PROT_EXEC)
	    flush_icache_range(page.start, page.start + PAGE_SIZE);
	r = read_kern(ctx, file, &page, sizeof(page));
    }
    if (r != sizeof(page)) goto err;

    if (k_mprotect(head->start,head->end - head->start, mmap_prot))
	printk("vmadump: thaw: mprotect failed. (ignoring)\n");
    return 0;

 err:
    if (r >= 0) r = -EIO;	/* map short reads to EIO */
    return r;
}

long vmadump_thaw_proc(struct vmadump_map_ctx *ctx,
		       struct file *file, struct pt_regs *regs) {
    int r;

    {
    struct vmadump_header header;

    /*--- First some sanity checking ---*/
    r = read_kern(ctx, file, &header, sizeof(header));
    if (r != sizeof(header)) {
	/*printk("vmadump: failed to read header: %d\n", r);*/
	return -EINVAL;
    }
    if (memcmp(header.magic, vmad_magic, sizeof(header.magic))) {
	/*printk("vmadump: invalid signature\n");*/
	return -EINVAL;
    }
    if (header.fmt_vers != VMAD_FMT_VERS) {
	printk(KERN_DEBUG "vmadump: dump version mistmatch. dump=%d; "
	       "kernel=%d\n", (int)header.fmt_vers, (int)VMAD_FMT_VERS);
	return -EINVAL;
    }
    if (header.arch != VMAD_ARCH) {
	printk(KERN_DEBUG "vmadump: architecture mismatch.\n");
	return -EINVAL;
    }

    if (header.major != ((LINUX_VERSION_CODE >> 16) & 0xFF) ||
	header.minor != ((LINUX_VERSION_CODE >> 8) & 0xFF)
#if STRICT_VERSION_CHECK
	||header.patch != (LINUX_VERSION_CODE & 0xFF)
#endif
	) {
	printk(KERN_DEBUG "vmadump: kernel version mismatch.\n");
	return -EINVAL;
    }
    }

    /* Ummm... Point of no-return is here.... maybe try to move this
     * down a bit? */

    /* Read our new comm */
    r = read_kern(ctx, file, current->comm, sizeof(current->comm));
    if (r != sizeof(current->comm)) goto bad_read;

    /*
     * CPU-specific register restore stuff
     *
     * Note that we're not presuming that our current regs pointer
     * points to anything even vaguely reasonable.  This is done to
     * support bproc type kernel threads that have never been user
     * processes before.
     */
    {
    struct pt_regs regtmp;

    r = read_kern(ctx, file, &regtmp, sizeof(regtmp));
    if (r != sizeof(regtmp)) goto bad_read;
#if defined(__i386__)
    {
    union i387_union     i387tmp;
    struct thread_struct threadtmp;
    char flag;

    /* Don't let the user pick bogo-segregs.  Restoring other values
     * will either lead us to fault while restoring or worse it might
     * allow users to do bad(tm) things in kernel space. */
    regtmp.xcs = __USER_CS;
    regtmp.xds = __USER_DS;
    regtmp.xes = __USER_DS;
    regtmp.xss = __USER_DS;
    /* fs and gs aren't represented in struct pt_regs... *hrm* */

    /* Only restore bottom 9 bits of eflags.  Restoring anything else
     * is bad bad mojo for security. (0x200 = interrupt enable) */
    regtmp.eflags = 0x200 | (regtmp.eflags & 0x000000FF);
    memcpy(regs, &regtmp, sizeof(regtmp));

    r = read_kern(ctx, file, &flag, sizeof(flag));
    if (r != sizeof(flag)) goto bad_read;

    if (flag) {
	/* Restore i386 FPU.
	 * - Clear TS so that we can do FPU ops.
	 * - Set PF_USEDFPU so the kernel knows that what's in the fpu right
	 *   now is of interest to us.
	 * - Read our dump data into the tss.
	 * - Manually restore from tss.
	 * - If the restore causes a GPF:
	 *     - Set TS so that FPU ops will trap.
	 *     - Clear used_math and PF_USEDFPU so the kernel will
	 *       give us a clean FPU when we trap.
	 */
	r = read_kern(ctx, file, &i387tmp, sizeof(i387tmp));
	if (r != sizeof(current->thread.i387)) goto bad_read;
	memcpy(&current->thread.i387, &i387tmp, sizeof(i387tmp));
	clts();
	current->used_math  = 1;
	current->flags     |= PF_USEDFPU;

	/* Invalid FPU states can blow us out of the water so we will do
	 * the restore here in such a way that we trap the fault if the
	 * restore fails.  This modeled after get_user and put_user. */
	if (cpu_has_fxsr) {
	    asm volatile
		("1: fxrstor %1               \n"
		 "2:                          \n"
		 ".section .fixup,\"ax\"      \n"
		 "3:  movl %2, %0             \n"
		 "    jmp 2b                  \n"
		 ".previous                   \n"
		 ".section __ex_table,\"a\"   \n"
		 "    .align 4                \n"
		 "    .long 1b, 3b            \n"
		 ".previous                   \n"
		 : "=r"(r)
		 : "m" (current->thread.i387.fxsave), "i"(-EFAULT));
	} else {
	    asm volatile
		("1: frstor %1                \n"
		 "2:                          \n"
		 ".section .fixup,\"ax\"      \n"
		 "3:  movl %2, %0             \n"
		 "    jmp 2b                  \n"
		 ".previous                   \n"
		 ".section __ex_table,\"a\"   \n"
		 "    .align 4                \n"
		 "    .long 1b, 3b            \n"
		 ".previous                   \n"
		 : "=r"(r)
		 : "m" (current->thread.i387.fsave), "i"(-EFAULT));
	}

	if (r == -EFAULT) {
	    /* Check for restore failure and bitch about it */
	    printk("vmadump: %d: FPU restore failure.\n", current->pid);
	    current->used_math  = 0;
	    current->flags     &= ~PF_USEDFPU;
	    stts();
	}
    } else {
	/* No FPU state for this process - setup for a clean FPU */
	current->used_math  = 0;
	current->flags     &= ~PF_USEDFPU;
	stts();
    }

    r = read_kern(ctx, file, threadtmp.debugreg, sizeof(threadtmp.debugreg));
    if (r != sizeof(threadtmp.debugreg)) goto bad_read;

    /* XXX FIX ME: RESTORE DEBUG INFORMATION ?? */
    }
#elif defined(__sparc__)
    regtmp.psr = regs->psr;	/* Don't restore PSR */
    memcpy(regs, &regtmp, sizeof(regtmp));
    /* no freely available docs = no fpu support */
#elif defined(__alpha__)
    {
	struct switch_stack *ss, sstmp;
	unsigned long usp;
	ss = ((struct switch_stack *)regs)-1;

	/* Read extra reg info for alpha */
	r = read_kern(ctx, file, &sstmp, sizeof(sstmp));
	if (r != sizeof(sstmp)) goto bad_read;
	r = read_kern(ctx, file, &usp, sizeof(usp));
	if (r != sizeof(usp)) goto bad_read;

	/* These are the tid-bits we don't want to restore */
	regtmp.ps  = 8;		/* Processor status 8 = user space */
	regtmp.hae = regs->hae;	/* XXX What is this reg? */
	sstmp.r26  = ss->r26;	/* Not actually the user r26(ra). That
				 * one is in struct pt_regs. */
	/* Ok, do restore */
	memcpy(regs, &regtmp, sizeof(regtmp));
	memcpy(ss, &sstmp, sizeof(sstmp)); /* Restores FP regs + saved regs */
	wrusp(usp);
    }
#elif defined(powerpc)
    {
 	/* Don't restore machine state register but set the PR bit
	 * since this is unpriviledged user space stuff we're
	 * restoring. */
	regtmp.msr = regs->msr | MSR_PR;
	memcpy(regs, &regtmp, sizeof(regtmp));

	r = read_kern(ctx, file, &current->thread.fpr,
		      sizeof(current->thread.fpr));
	if (r != sizeof(current->thread.fpr)) goto bad_read;

	r = read_kern(ctx, file, &current->thread.fpscr,
		      sizeof(current->thread.fpscr));
	if (r != sizeof(current->thread.fpscr)) goto bad_read;

#ifdef CONFIG_ALTIVEC
	/* Restore Altivec */
	if (regs->msr & MSR_VEC)
	    giveup_altivec(current);
	r = read_kern(ctx, file, &current->thread.vr,
		      sizeof(current->thread.vr));
	if (r != sizeof(current->thread.vr)) goto bad_read;

	r = read_kern(ctx, file, &current->thread.vscr,
		      sizeof(current->thread.vscr));
	if (r != sizeof(current->thread.vscr)) goto bad_read;
#endif
    }
#else
#error No processor specific state restore defined.
#endif
    }

    /*--- Signal information ---------------------------------------*/
    {
    int i;
    sigset_t             sig_blocked;
    struct k_sigaction   sig_action;
	
    /* Install set of blocked signals */
    r = read_kern(ctx, file, &sig_blocked, sizeof(sig_blocked));
    if (r != sizeof(sig_blocked)) goto bad_read;

    sigdelsetmask(&sig_blocked, sigmask(SIGKILL)|sigmask(SIGSTOP));
    spin_lock_irq(&current->sigmask_lock);
    flush_signals(current);	/* Necessary? */
    memcpy(&current->blocked, &sig_blocked, sizeof(sig_blocked));
    recalc_sigpending(current);
    spin_unlock_irq(&current->sigmask_lock);

    for (i=0; i < _NSIG; i++) {
	r = read_kern(ctx, file, &sig_action, sizeof(sig_action));
	if (r != sizeof(sig_action)) goto bad_read;
	/* Security: Make sure the user doesn't block or handle the
	 * unblockable signals. */
	if (i == SIGSTOP-1 || i == SIGKILL-1) {
	    sig_action.sa.sa_handler = SIG_DFL;
	    sig_action.sa.sa_flags   = 0;
	    sigemptyset(&sig_action.sa.sa_mask);
	}
	sigdelsetmask(&sig_action.sa.sa_mask,
		      sigmask(SIGKILL)|sigmask(SIGSTOP));

	/* Install one sigaction */
	spin_lock_irq(&current->sigmask_lock);
	flush_signals(current);	/* Necessary? */
	memcpy(&current->sig->action[i], &sig_action, sizeof(sig_action));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
    }
    }

    /*--- Memory map meta data -------------------------------------*/
    {
    struct mm_struct *mm;
    struct vm_area_struct *map;
    struct vmadump_mm_info mm_info;
    struct vmadump_vma_header mapheader;

    mm = current->mm;

    r = read_kern(ctx, file, &mm_info, sizeof(mm_info));
    if (r != sizeof(mm_info)) { goto bad_read; }

    /* Purge current maps - I'm sure there's a way to keep theses around
     * incase creation of the new ones fails in some unfortunate way... */
    while(mm->mmap) {
	map = mm->mmap;
	r = vmad_do_munmap(mm, map->vm_start, map->vm_end - map->vm_start);
	if (r) {
	    printk("do_munmap(%lu, %lu) = %d\n",
		   map->vm_start, map->vm_end-map->vm_start,r);
	}
    }

    /* Load new map data */
    r = read_kern(ctx, file, &mapheader, sizeof(mapheader));
    while (r == sizeof(mapheader) &&
	   (mapheader.start != ~0 || mapheader.end != ~0)) {
	if ((r = load_map(ctx, file, &mapheader))) goto bad_read;
	r = read_kern(ctx, file, &mapheader, sizeof(mapheader));
    }
    if (r != sizeof(mapheader)) goto bad_read;    

    down_write(&current->mm->mmap_sem);
    mm->start_code = mm_info.start_code;
    mm->end_code   = mm_info.end_code;
    mm->start_data = mm_info.start_data;
    mm->end_data   = mm_info.end_data;
    mm->start_brk  = mm_info.start_brk;
    mm->brk        = mm_info.brk;
    mm->start_stack= mm_info.start_stack;
    mm->arg_start  = mm_info.arg_start;
    mm->arg_end    = mm_info.arg_end;
    mm->env_start  = mm_info.env_start;
    mm->env_end    = mm_info.env_end;
    up_write(&current->mm->mmap_sem);
    }

    /*--- Call external thaw hooks ---------------------------------*/
    r = do_thaw_hooks(ctx, file, regs);
    if (r) goto bad_read;

    return 0;

 bad_read:
    if (r >= 0) r = -EIO;
    return r;
}

/*--------------------------------------------------------------------
 *  Process "freezing" routines
 *------------------------------------------------------------------*/
/* This routine checks if a page from a filemap has been copied via
 * copy on write.  Basically, this is just checking to see if the page
 * is still a member of the map or not.  Note this this should not end
 * up copying random things from VM_IO regions. */
static
int addr_copied(struct mm_struct *mm, unsigned long addr) {
    pgd_t *pgd;
    pmd_t *pmd;
    pte_t *pte;
    struct page *pg;
    int ret;

    spin_lock(&mm->page_table_lock);
    pgd = pgd_offset(mm, addr);
    if (pgd_none(*pgd)) { ret = 0; goto out; }
    pmd = pmd_offset(pgd, addr);
    if (pmd_none(*pmd)) { ret = 0; goto out; }
    pte = pte_offset(pmd, addr);
    if (pte_present(*pte)) {
	pg = pte_page(*pte);
	ret = !pg->mapping;
	/* I think this was paranoia:
	   ret = !pg->mapping || pg->mapping !=
           map->vm_file->f_dentry->d_inode->i_mapping; */
    } else
	/* pte_none is false for a swapped (written) page */
	ret = !pte_none(*pte);
 out:
    spin_unlock(&mm->page_table_lock);
    return ret;
}

/* This is the version for working on a region that is a file map.  In
 * this case we need to fault the page in to check for zero.  This
 * isn't a big deal since we'll be faulting in for sending anyway if
 * it's not.  */
static
int addr_nonzero_file(struct mm_struct *mm, unsigned long addr) {
    int i;
    unsigned long val;

    /* Simple zero check */
    for (i=0; i < (PAGE_SIZE/sizeof(long)); i++) {
	get_user(val, (((long*)addr)+i));
	if (val) return 1;
    }
    return 0;
}

/* This version is for use on regions which are *NOT* file maps.  Here
 * we look at the page tables to see if a page is zero.  If it's never
 * been faulted in, we know it's zero - and we don't fault it in while
 * checking for this. */
static
int addr_nonzero(struct mm_struct *mm, unsigned long addr) {
    int i;
    pgd_t *pgd;
    pmd_t *pmd;
    pte_t *pte;
    unsigned long val;

    spin_lock(&mm->page_table_lock);
    pgd = pgd_offset(mm, addr);
    if (pgd_none(*pgd)) goto out_zero;
    pmd = pmd_offset(pgd, addr);
    if (pmd_none(*pmd)) goto out_zero;
    pte = pte_offset(pmd, addr);
    if (pte_none(*pte)) goto out_zero;
    spin_unlock(&mm->page_table_lock);

    /* Ok, the page could be non-zero - check it... */
    for (i=0; i < (PAGE_SIZE/sizeof(long)); i++) {
	get_user(val, (((long*)addr)+i));
	if (val) return 1;
    }
    return 0;

 out_zero:
    spin_unlock(&mm->page_table_lock);
    return 0;
}

static
int store_page_list_eof(struct vmadump_map_ctx *ctx, struct file *file) {
    int r;
    struct vmadump_page_header header;
    header.start = ~0UL;
    r = write_kern(ctx, file, &header, sizeof(header));
    if (r != sizeof(header)) r = -EIO;
    return r;
}

static
int store_page_list(struct vmadump_map_ctx *ctx, struct file *file,
		    unsigned long start, unsigned long end,
		    int (*pfunc)(struct mm_struct *mm, unsigned long)) {
    int r, bytes = 0;
    unsigned long addr;
    struct vmadump_page_header header;
    for (addr = start; addr < end; addr += PAGE_SIZE) {
	if (pfunc(current->mm, addr)) {
	    header.start = addr;
	    r = write_kern(ctx, file, &header, sizeof(header));
	    if (r != sizeof(header)) { r = -EIO; return r; }
	    bytes += r;
	    r = write_user(ctx, file, (void *)addr, PAGE_SIZE);
	    if (r != PAGE_SIZE) { r = -EIO; return r; }
	    bytes += r;
	}
    }
    r = store_page_list_eof(ctx, file);
    if (r < 0) return r;
    bytes += r;
    return bytes;
}

static
int store_map(struct vmadump_map_ctx *ctx, struct file *file,
	      struct vm_area_struct *map, int flags) {
    int bytes;
    struct vmadump_vma_header head;
    char *filename=0;
    char *buffer = 0;
    int r;
    unsigned long start, end;
    int isfilemap;

    head.start   = map->vm_start;
    head.end     = map->vm_end;
    head.flags   = map->vm_flags;
    head.namelen = 0;
    head.offset  = map->vm_pgoff << PAGE_SHIFT; /* XXX LFS! */

    /* Decide Whether or not we're gonna store the map's contents or
     * a reference to the file they came from */
    if (map->vm_file) {
	buffer = (char *) get_free_page(GFP_KERNEL);
	if (!buffer) { return -ENOMEM; }
	if (ctx && ctx->map_name)
	    filename = ctx->map_name(ctx, map->vm_file, buffer, PAGE_SIZE);
	else
	    filename = default_map_name(map->vm_file, buffer, PAGE_SIZE);
	head.namelen = strlen(filename);

	if (map->vm_flags & VM_IO) {
	    /* Region is an IO map. */
	    
	    /* Never store the contents of a VM_IO region */
	} else if (map->vm_flags & VM_EXECUTABLE) {
	    /* Region is an executable */
	    if (flags & VMAD_DUMP_EXEC)
		head.namelen = 0;
	} else if (is_library(filename)) {
	    /* Region is a library */
	    if (flags & VMAD_DUMP_LIBS)
		head.namelen = 0;
	} else {
	    /* Region is something else */
	    if (flags & VMAD_DUMP_OTHER)
		head.namelen=0;
	}
    }

    start     = map->vm_start;
    end       = map->vm_end;
    isfilemap = (map->vm_file != 0);
    /* Release the mm_sem here to avoid deadlocks with page faults and
     * write locks that may happen during the writes.  (We can't use
     * the "map" pointer beyond this point. */
    up_read(&current->mm->mmap_sem);

    /* Spit out the section header */
    r = write_kern(ctx, file, &head, sizeof(head));
    if (r != sizeof(head)) goto err;
    bytes = r;

    if (head.namelen) {
	/* Store the filename */
	r = write_kern(ctx, file, filename, head.namelen);
	if (r != head.namelen) goto err;
	bytes += r;
	r = store_page_list(ctx, file, start, end, addr_copied);
	if (r < 0) goto err;
	bytes += r;
    } else {
	/* Store the contents of the VMA as defined by start, end */
	r = store_page_list(ctx, file, start, end,
			    isfilemap ? addr_nonzero_file : addr_nonzero);
	if (r < 0) goto err;
	bytes += r;
    }
    if (buffer)   free_page((long)buffer);
    down_read(&current->mm->mmap_sem);
    return bytes;

 err:
    if (r >= 0) r = -EIO;	/* Map short writes to EIO */
    if (buffer)   free_page((long)buffer);
    down_read(&current->mm->mmap_sem);
    return r;
}

long vmadump_freeze_proc(struct vmadump_map_ctx *ctx, struct file *file,
			struct pt_regs *regs, int flags) {
    long r, bytes=0;
    static struct vmadump_header header ={VMAD_MAGIC, VMAD_FMT_VERS, VMAD_ARCH,
					  (LINUX_VERSION_CODE >> 16) & 0xFF,
					  (LINUX_VERSION_CODE >> 8) & 0xFF,
					  LINUX_VERSION_CODE & 0xFF };

    /*--- Write out the file header ---*/
    r = write_kern(ctx, file, &header, sizeof(header));
    if (r != sizeof(header)) goto err;
    bytes += r;

    r = write_kern(ctx, file, current->comm, sizeof(current->comm));
    if (r != sizeof(current->comm)) goto err;
    bytes += r;

    /*--- CPU State Information ------------------------------------*/
    r = write_kern(ctx, file, regs, sizeof(*regs));
    if (r != sizeof(*regs)) goto err;
    bytes += r;

#if defined(__i386__)
    {
	struct user_i387_struct dummy;
	union i387_union i387tmp;
	char flag;

	/* The i386's FPU state isn't included in struct pt_regs, so
	 * save it seperately here. */

	/* Explanation of flags:
	 * used_math  =  1 iff process has used FPU
	 * PF_USEDFPU =  1 iff process has used math since last context switch
	 * TS         =  1 iff task switched, i.e. fpu ops will cause trap
	 */

	/* If we haven't touched the FPU, make like we have so that we
	 * have something to save.  First initialize it and then store
	 * it in our thread (unlazy does this.) */
	if (current->used_math) {
	    /* If the FPU has been used, try to save the state */
	    flag = 1;
	    r = write_kern(ctx, file, &flag, sizeof(flag));
	    if (r != sizeof(flag)) goto err;

	    if (current->flags & PF_USEDFPU) {
		/* hack here, unlazy in not available from to modules
		 * so call dump_fpu which has unlazy as a side effect.
		 * *sigh* */
		dump_fpu(0, &dummy);
	    }
	    memcpy(&i387tmp, &current->thread.i387, sizeof(i387tmp));
	    r = write_kern(ctx, file, &i387tmp, sizeof(i387tmp));
	    if (r != sizeof(i387tmp)) goto err;
	    bytes += r;
	} else {
	    /* No math used */
	    flag = 0;
	    r = write_kern(ctx, file, &flag, sizeof(flag));
	    if (r != sizeof(flag)) goto err;
	    bytes += r;
	}

	/* Store debugging state */
	r = write_kern(ctx, file, &current->thread.debugreg,
		      sizeof(current->thread.debugreg));
	if (r != sizeof(current->thread.debugreg)) goto err;
	bytes += r;
    }
#elif defined(__alpha__)
    /* Alpha has more registers than what's in struct ptregs to save
     * The 'switch_stack' structure contains the rest of the integer
     * registers and the floating point registers.  Then we still have
     * to go after the user's stack pointer via the PAL code. */
    {
	struct switch_stack *ss;
	unsigned long usp;
	/* Some pointer voodoo... the switch_stack structure got
	 * pushed on top of the struct pt_regs.. */
	ss = ((struct switch_stack *) regs)-1;
	usp = rdusp();

	r = write_kern(ctx, file, ss, sizeof(*ss));
	if (r != sizeof(*ss)) goto err;
	bytes += r;

	r = write_kern(ctx, file, &usp, sizeof(usp));
	if (r != sizeof(usp)) goto err;
	bytes += r;
    }
#elif defined(powerpc)
    {
	/* Floating point regs */
	if (regs->msr & MSR_FP)
	    giveup_fpu(current);
	r = write_kern(ctx, file, &current->thread.fpr,
		       sizeof(current->thread.fpr));
	if (r != sizeof(current->thread.fpr)) goto err;
	bytes += r;

	r = write_kern(ctx, file, &current->thread.fpscr,
		       sizeof(current->thread.fpscr));
	if (r != sizeof(current->thread.fpscr)) goto err;
	bytes += r;
#ifdef CONFIG_ALTIVEC
	/* XXX I really need to find out if this is right */
	if (regs->msr & MSR_VEC)
	    giveup_altivec(current);
	r = write_kern(ctx, file, &current->thread.vr,
		       sizeof(current->thread.vr));
	if (r != sizeof(current->thread.vr)) goto err;
	bytes += r;

	r = write_kern(ctx, file, &current->thread.vscr,
		       sizeof(current->thread.vscr));
	if (r != sizeof(current->thread.vscr)) goto err;
	bytes += r;
#endif
    }
#else
#error No processor specific state save defined.
#endif

    /*--- Signal information ---------------------------------------*/
    {
    int i;
    sigset_t             sig_blocked;
    struct k_sigaction   sig_action;

    spin_lock_irq(&current->sigmask_lock);
    memcpy(&sig_blocked, &current->blocked, sizeof(sig_blocked));
    spin_unlock_irq(&current->sigmask_lock);

    r = write_kern(ctx, file, &sig_blocked, sizeof(sig_blocked));
    if (r != sizeof(sig_blocked)) goto err1;
    bytes += r;
    for (i=0; i < _NSIG; i++) {
	spin_lock_irq(&current->sigmask_lock);
	memcpy(&sig_action, &current->sig->action[i], sizeof(sig_action));
	spin_unlock_irq(&current->sigmask_lock);
	
	r = write_kern(ctx, file, &sig_action, sizeof(sig_action));
	if (r != sizeof(sig_action)) goto err1;
	bytes += r;
    }
    }
    
    /*--- Memory Information ---------------------------------------*/
    {
    struct vm_area_struct     *map, *next_map;
    struct vmadump_mm_info     mm_info;
    struct mm_struct          *mm = current->mm;
    struct vmadump_vma_header  term;

    down_read(&mm->mmap_sem);
    mm_info.start_code  = mm->start_code;
    mm_info.end_code    = mm->end_code;
    mm_info.start_data  = mm->start_data;
    mm_info.end_data    = mm->end_data;
    mm_info.start_brk   = mm->start_brk;
    mm_info.brk         = mm->brk;
    mm_info.start_stack = mm->start_stack;
    mm_info.arg_start   = mm->arg_start;
    mm_info.arg_end     = mm->arg_end;
    mm_info.env_start   = mm->env_start;
    mm_info.env_end     = mm->env_end;
    up_read(&mm->mmap_sem);

    r = write_kern(ctx, file, &mm_info, sizeof(mm_info));
    if (r != sizeof(mm_info)) goto err;
    bytes += r;

    down_read(&mm->mmap_sem);
    next_map = mm->mmap;
    while (next_map) {
	/* Scan forward till we find the map we're looking for.  We
	 * have to do it this way because store_map needs to release
	 * the mmap_sem. */
	for (map = mm->mmap; map && map != next_map; map = map->vm_next);
	if (!map) break;
	next_map = map->vm_next;
	r = store_map(ctx, file, map, flags);
	if (r < 0) {
	    up_read(&mm->mmap_sem);
	    goto err;
	}
	bytes += r;
    }
    up_read(&mm->mmap_sem);
    
    /* Terminate maps list */
    term.start = term.end = ~0L;
    r = write_kern(ctx, file, &term, sizeof(term));
    if (r != sizeof(term)) goto err;
    bytes += r;
    }

    /*--- Call freeze hooks ----------------------------------------*/
    r = do_freeze_hooks(ctx, file, regs);
    if (r < 0) goto err;
    bytes += r;

    return bytes;

 err1:
    spin_unlock_irq(&current->sigmask_lock);
 err:
    if (r >= 0) r = -EIO;	/* Map short writes to EIO */
    return r;
}

/*--------------------------------------------------------------------
 * syscall interface
 *------------------------------------------------------------------*/
/* These tell us how to get syscall arguments out of a struct pt_regs */
#if defined(__i386__)
#define reg_arg0(r)      ((r)->ebx)
#define reg_arg1(r)      ((r)->ecx)
#define reg_arg2(r)      ((r)->edx)
#elif defined(__sparc__)
#define reg_arg0(r)      ((r)->u_regs[((r)->u_regs[UREG_G1]==0?1:0)+UREG_I0])
#define reg_arg1(r)      ((r)->u_regs[((r)->u_regs[UREG_G1]==0?1:0)+UREG_I1])
#define reg_arg2(r)      ((r)->u_regs[((r)->u_regs[UREG_G1]==0?1:0)+UREG_I2])
#elif defined(__alpha__)
#define reg_arg0(r)      ((r)->trap_a0)
#define reg_arg1(r)      ((r)->trap_a1)
#define reg_arg2(r)      ((r)->trap_a2)
#elif defined(powerpc)
#define reg_arg0(r)      ((r)->gpr[3])
#define reg_arg1(r)      ((r)->gpr[4])
#define reg_arg2(r)      ((r)->gpr[5])
#else
#error reg_arg[0-2] needs to be defined for this arch
#endif

long do_vmadump(struct pt_regs *regs) {
    long retval;
    int op    = reg_arg0(regs);
    int fd    = reg_arg1(regs);
    int flags = reg_arg2(regs);
    struct file *file;

    MOD_INC_USE_COUNT;
    switch (op) {
    case VMAD_DO_DUMP:
	if ((file = fget(fd))) {
	    retval = vmadump_freeze_proc(0, file, regs, flags);
	    fput(file);
	} else
	    retval = -EBADF;
	break;
    case VMAD_DO_UNDUMP:
	if ((file = fget(fd))) {
	    retval = vmadump_thaw_proc(0, file, regs);
	    fput(file);
	} else
	    retval = -EBADF;

	/* Un-dump is a whole lot like exec() */
	if (retval == 0) {
	    if (current->euid == current->uid && current->egid == current->gid)
		current->mm->dumpable = 1;
	    current->did_exec = 1;
	    current->self_exec_id++;
	    if (current->ptrace & PT_PTRACED)
                send_sig(SIGTRAP, current, 0);
	}
	break;
    case VMAD_DO_EXECDUMP: {
	struct vmadump_execdump_args args;
	char * filename;
	retval = -EFAULT;
	if (copy_from_user(&args, (void *) reg_arg1(regs), sizeof(args)))
	    break;

	/* Check permissions on the binary image. */
	retval = k_access(args.arg0, 4);
	if (retval)
	    break;

	filename = getname(args.arg0);
	retval = PTR_ERR(filename);
	if (IS_ERR(filename)) break;

	file = fget(args.fd);
	if (!file) {
	    retval = -EBADF;
	    putname(filename);
	    break;
	}

	retval = do_execve(filename, (char **)args.argv,
			   (char **)args.envp, regs);
	putname(filename);
	if (retval) {
	    fput(file);
	    break;
	}

	retval = vmadump_freeze_proc(0, file, regs, args.flags);
	fput(file);
	if (retval > 0) retval = 0;
	MOD_DEC_USE_COUNT;
	k_exit(retval);		/* Ok, we're done... */
	/* NOT REACHED */
    } break;

    case VMAD_LIB_CLEAR:
    case VMAD_LIB_ADD:
    case VMAD_LIB_DEL:
    case VMAD_LIB_LIST:
    case VMAD_LIB_SIZE:
	lock_kernel();		/* very very lazy... */
	retval = do_lib_op(op, (char *) reg_arg1(regs), reg_arg2(regs));
	unlock_kernel();
	break;

    default:
	retval = -EINVAL;
    }
    MOD_DEC_USE_COUNT;
    return retval;
}

/*--------------------------------------------------------------------
 * Stubs for syscall entry on various architectures
 *------------------------------------------------------------------*/
#ifdef __i386__
static
int sys_vmadump(struct pt_regs regs) {
    return do_vmadump(&regs);
}
#endif

#ifdef __alpha__
#define SWITCH_STACK_SIZE "320"
int sys_vmadump(void);
/* do_switch_stack/undo_switch_stack stolen from arch/alpha/kernel/entry.S */
asm(
".align 3			\n"
".ent	save_switch_stack	\n"
"save_switch_stack:		\n"
"	lda	$30,-"SWITCH_STACK_SIZE"($30)	\n"
"	stq	$9,0($30)	\n"
"	stq	$10,8($30)	\n"
"	stq	$11,16($30)	\n"
"	stq	$12,24($30)	\n"
"	stq	$13,32($30)	\n"
"	stq	$14,40($30)	\n"
"	stq	$15,48($30)	\n"
"	stq	$26,56($30)	\n"
"	stt	$f0,64($30)	\n"
"	stt	$f1,72($30)	\n"
"	stt	$f2,80($30)	\n"
"	stt	$f3,88($30)	\n"
"	stt	$f4,96($30)	\n"
"	stt	$f5,104($30)	\n"
"	stt	$f6,112($30)	\n"
"	stt	$f7,120($30)	\n"
"	stt	$f8,128($30)	\n"
"	stt	$f9,136($30)	\n"
"	stt	$f10,144($30)	\n"
"	stt	$f11,152($30)	\n"
"	stt	$f12,160($30)	\n"
"	stt	$f13,168($30)	\n"
"	stt	$f14,176($30)	\n"
"	stt	$f15,184($30)	\n"
"	stt	$f16,192($30)	\n"
"	stt	$f17,200($30)	\n"
"	stt	$f18,208($30)	\n"
"	stt	$f19,216($30)	\n"
"	stt	$f20,224($30)	\n"
"	stt	$f21,232($30)	\n"
"	stt	$f22,240($30)	\n"
"	stt	$f23,248($30)	\n"
"	stt	$f24,256($30)	\n"
"	stt	$f25,264($30)	\n"
"	stt	$f26,272($30)	\n"
"	stt	$f27,280($30)	\n"
"	mf_fpcr	$f0		\n"
"	stt	$f28,288($30)	\n"
"	stt	$f29,296($30)	\n"
"	stt	$f30,304($30)	\n"
"	stt	$f0,312($30)	\n"
"	ldt	$f0,64($30)	\n"
"	ret	$31,($1),1	\n"
".end save_switch_stack		\n"
"\n"
".align 3			\n"
".ent	restore_switch_stack	\n"
"restore_switch_stack:		\n"
"	ldq	$9,0($30)	\n"
"	ldq	$10,8($30)	\n"
"	ldq	$11,16($30)	\n"
"	ldq	$12,24($30)	\n"
"	ldq	$13,32($30)	\n"
"	ldq	$14,40($30)	\n"
"	ldq	$15,48($30)	\n"
"	ldq	$26,56($30)	\n"
"	ldt	$f30,312($30)	# get saved fpcr	\n"
"	ldt	$f0,64($30)	\n"
"	ldt	$f1,72($30)	\n"
"	ldt	$f2,80($30)	\n"
"	ldt	$f3,88($30)	\n"
"	mt_fpcr	$f30		# install saved fpcr	\n"
"	ldt	$f4,96($30)	\n"
"	ldt	$f5,104($30)	\n"
"	ldt	$f6,112($30)	\n"
"	ldt	$f7,120($30)	\n"
"	ldt	$f8,128($30)	\n"
"	ldt	$f9,136($30)	\n"
"	ldt	$f10,144($30)	\n"
"	ldt	$f11,152($30)	\n"
"	ldt	$f12,160($30)	\n"
"	ldt	$f13,168($30)	\n"
"	ldt	$f14,176($30)	\n"
"	ldt	$f15,184($30)	\n"
"	ldt	$f16,192($30)	\n"
"	ldt	$f17,200($30)	\n"
"	ldt	$f18,208($30)	\n"
"	ldt	$f19,216($30)	\n"
"	ldt	$f20,224($30)	\n"
"	ldt	$f21,232($30)	\n"
"	ldt	$f22,240($30)	\n"
"	ldt	$f23,248($30)	\n"
"	ldt	$f24,256($30)	\n"
"	ldt	$f25,264($30)	\n"
"	ldt	$f26,272($30)	\n"
"	ldt	$f27,280($30)	\n"
"	ldt	$f28,288($30)	\n"
"	ldt	$f29,296($30)	\n"
"	ldt	$f30,304($30)	\n"
"	lda	$30,"SWITCH_STACK_SIZE"($30)	\n"
"	ret	$31,($1),1	\n"
".end restore_switch_stack	\n"
"				\n"
".align 3			\n"
".globl  sys_vmadump		\n"
".ent    sys_vmadump		\n"
"sys_vmadump:			\n"
"        ldgp    $29,0($27)	\n"
"        bsr     $1,save_switch_stack	\n"
"        lda     $16,"SWITCH_STACK_SIZE"($30)	\n"
"        jsr     $26,do_vmadump	\n"
"        bsr     $1,restore_switch_stack	\n"
"        ret     $31,($26),1	\n"
".end    sys_vmadump		\n"
);
#endif

#ifdef __sparc__
int sys_vmadump(void);
#define STRING_1(x) #x
#define STRING(x) STRING_1(x)
#define LABEL(x) STRING(C_LABEL(x))
asm(
"	.align	4	\n"
"	.globl	"LABEL(sys_vmadump)"	\n"
LABEL(sys_vmadump)":	\n"
"	mov	%o7, %l5	\n"
"	add	%sp, 0x40, %o0	\n"
"	call	"LABEL(do_vmadump)"	\n"
"	 mov	%l5, %o7");
#endif

#ifdef powerpc
int sys_vmadump(int p1, int p2, int p3, int p4, int p5, int p6,
		struct pt_regs *regs) {
    return do_vmadump(regs);
}
#endif

/*--------------------------------------------------------------------
 *  New binary format code
 *------------------------------------------------------------------*/
#ifdef HAVE_BINFMT_VMADUMP

static
int load_vmadump(struct linux_binprm *bprm, struct pt_regs *regs) {
    int retval;
    struct vmadump_header *header;

    header = (struct vmadump_header *) bprm->buf;
    if (memcmp(header->magic, vmad_magic, sizeof(header->magic)) != 0  || 
	header->fmt_vers != VMAD_FMT_VERS ||
	header->major != ((LINUX_VERSION_CODE >> 16) & 0xFF) ||
	header->minor != ((LINUX_VERSION_CODE >> 8) & 0xFF)
#if STRICT_VERSION_CHECK
	|| header->patch != (LINUX_VERSION_CODE & 0xFF)
#endif
	)
	return -ENOEXEC;

    if (!bprm->file->f_op || !bprm->file->f_op->mmap)
	return -ENOEXEC;

    retval = vmadump_thaw_proc(0, bprm->file, regs);
    if (retval == 0) {
	if (current->euid == current->uid && current->egid == current->gid) 
	    current->mm->dumpable = 1;
	current->did_exec = 1;
	current->self_exec_id++;
        if (current->exec_domain && current->exec_domain->module)
	    __MOD_DEC_USE_COUNT(current->exec_domain->module);
        if (current->binfmt && current->binfmt->module)
	    __MOD_DEC_USE_COUNT(current->binfmt->module);
	current->exec_domain = 0;
	current->binfmt = 0;
    }
    return retval;
}

struct linux_binfmt vmadump_fmt = {
    0,THIS_MODULE,load_vmadump,0,0,0
};
#endif

static
char *get_rev(char *revstr) {
    char *ch; if ((ch=strchr(revstr+11,' ')))*ch = 0;
    return revstr+11;
}

static void * old_sys_call;
int init_module(void) {
    printk(KERN_INFO "vmadump: %s Erik Hendriks "
	   "<erik@hendriks.cx>\n", get_rev("$Revision: 1.7 $"));

    init_rwsem(&hook_lock);

    /* We can't get an upper bound on the number of syscalls so be
     * careful here! */
    if (syscall_num < 0) return -EINVAL;

#ifdef HAVE_BINFMT_VMADUMP
    register_binfmt(&vmadump_fmt);
#endif

    if (syscall_num) {
	printk(KERN_INFO "vmadump: syscall number %d\n", syscall_num);
	old_sys_call = sys_call_table[syscall_num];
	sys_call_table[syscall_num] = sys_vmadump;
    }
    return 0;
}

void cleanup_module(void) {
    liblist_clear();
    if (syscall_num) {
	sys_call_table[syscall_num] = old_sys_call;
    }
#ifdef HAVE_BINFMT_VMADUMP
    unregister_binfmt(&vmadump_fmt);
#endif
}

EXPORT_SYMBOL(vmadump_freeze_proc);
EXPORT_SYMBOL(vmadump_thaw_proc);
EXPORT_SYMBOL(do_vmadump);

EXPORT_SYMBOL(vmadump_add_hook);
EXPORT_SYMBOL(vmadump_del_hook);
EXPORT_SYMBOL(vmadump_read_u);
EXPORT_SYMBOL(vmadump_read_k);
EXPORT_SYMBOL(vmadump_write_u);
EXPORT_SYMBOL(vmadump_write_k);
#ifdef __WITH__CHPOX__
EXPORT_SYMBOL(vmadump_fmt);
#endif
/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
